# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from abc import ABC
from collections import OrderedDict
from contextlib import ExitStack, contextmanager
from numbers import Number
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
from compiler_config.config import CompilerConfig

from qat.backend.base import BaseBackend
from qat.backend.codegen_base import DfsTraversal
from qat.backend.passes.purr.analysis import (
    CFGPass,
    CFGResult,
    ReadWriteResult,
    ScopingResult,
    TILegalisationPass,
)
from qat.backend.qblox.config.constants import Constants
from qat.backend.qblox.config.specification import SequencerConfig
from qat.backend.qblox.execution import QbloxExecutable, QbloxPackage
from qat.backend.qblox.ir import Opcode, SequenceBuilder
from qat.backend.qblox.passes.analysis import (
    AllocationManager,
    BindingResult,
    IterBound,
    PreCodegenPass,
    PreCodegenResult,
    QbloxLegalisationPass,
    TriageResult,
)
from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import InvokerMixin, PassManager
from qat.core.result_base import ResultManager
from qat.executables import Executable
from qat.purr.backends.utilities import evaluate_shape
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, PulseShapeType
from qat.purr.compiler.execution import DeviceInjectors, SweepIterator
from qat.purr.compiler.instructions import (
    Acquire,
    CustomPulse,
    Delay,
    DeviceUpdate,
    FrequencyShift,
    Id,
    Instruction,
    MeasurePulse,
    PhaseReset,
    PhaseShift,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    Pulse,
    QuantumInstruction,
    Repeat,
    Sweep,
    Synchronize,
    Variable,
    Waveform,
    calculate_duration,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class AbstractContext(ABC):
    def __init__(self, alloc_mgr: AllocationManager = None):
        self.alloc_mgr = alloc_mgr or AllocationManager()
        self.sequence_builder = SequenceBuilder()
        self.sequencer_config = SequencerConfig()

        self._num_hw_avg = 1  # Technically disabled
        self._wf_memory: int = Constants.MAX_SAMPLE_SIZE_WAVEFORMS
        self._wf_index: int = 0
        self._acq_index: int = 0
        self._wgt_index: int = 0

        self._durations: List[Union[float, str]] = []
        self._frequency: float = 0.0  # Tracks frequency shifts on the target
        self._phases: List[Union[float, str]] = []
        self._timeline: np.ndarray = np.empty(0, dtype=complex)

    @property
    def durations(self):
        return self._durations

    @property
    def duration(self):
        if all((isinstance(d, Number) for d in self._durations)):
            return sum(self._durations)

        ValueError("Cannot determine duration statically in dynamic settings")

    @property
    def phase(self):
        if all((isinstance(p, Number) for p in self._phases)):
            return sum(self._phases)

        ValueError("Cannot determine phase statically in dynamic settings")

    def create_package(self):
        sequence = self.sequence_builder.build()
        return QbloxPackage(sequence, self.sequencer_config, self._timeline)

    def is_empty(self):
        """
        Masks away yet-to-be-supported second-state, cancellation, and cross cancellation targets
        This is temporary and criteria will change with more features coming in
        """
        return not (
            self.sequence_builder.waveforms
            or self.sequence_builder.acquisitions
            or [
                inst
                for inst in self.sequence_builder.q1asm_instructions
                if inst.opcode
                in [
                    Opcode.PLAY,
                    Opcode.SET_AWG_OFFSET,
                    Opcode.SET_AWG_GAIN,
                    Opcode.ACQUIRE,
                    Opcode.ACQUIRE_TTL,
                    Opcode.ACQUIRE_WEIGHED,
                ]
            ]
        )

    def clear(self):
        self.sequence_builder.waveforms.clear()
        self.sequence_builder.acquisitions.clear()
        self.sequence_builder.weights.clear()
        self.sequence_builder.q1asm_instructions.clear()

    def wait_imm(self, duration: int):
        """
        Waits for `duration` nanoseconds expressed as an immediate.

        `duration` must be positive
        """

        if duration <= 0:
            log.debug(f"Expected positive duration, but got {duration}")
            return

        quotient = duration // Constants.MAX_WAIT_TIME
        remainder = duration % Constants.MAX_WAIT_TIME

        if remainder < Constants.GRID_TIME:
            log.debug(f"Rounding up {remainder} ns to {Constants.GRID_TIME} ns")
            remainder = Constants.GRID_TIME
            duration = quotient * Constants.MAX_WAIT_TIME + remainder

        if quotient >= Constants.LOOP_UNROLL_THRESHOLD:
            with self._loop("wait_label", quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
        elif quotient >= 1:
            for _ in range(quotient):
                self.sequence_builder.wait(Constants.MAX_WAIT_TIME)

        self.sequence_builder.wait(remainder)

    def wait_reg(self, duration: str):
        """
        A mere `wait RX` in general has undefined runtime behaviour.

        This is a useful helper to dynamically wait for a `duration` nanoseconds expressed as a register.
        `iter_reg` is a short-lived register only used as an interator. Customer is responsible for (de)allocation.
        """

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")
        remainder = self.alloc_mgr.label_gen("remainder")
        round_up = self.alloc_mgr.label_gen("round_up")

        with self.alloc_mgr.reg_borrow("tmp") as iter_reg:
            if duration != iter_reg:
                self.sequence_builder.move(
                    duration,
                    iter_reg,
                    f"Avoids cluttering {duration} as it's likely used as an accumulator",
                )
                self.sequence_builder.nop()

            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, round_up)
            self.sequence_builder.jlt(iter_reg, Constants.MAX_WAIT_TIME, remainder)
            self.sequence_builder.label(batch_enter)
            self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
            self.sequence_builder.sub(iter_reg, Constants.MAX_WAIT_TIME, iter_reg)
            self.sequence_builder.nop()
            self.sequence_builder.jge(iter_reg, Constants.MAX_WAIT_TIME, batch_enter)
            self.sequence_builder.jge(iter_reg, Constants.GRID_TIME, remainder)
            self.sequence_builder.label(round_up)
            self.sequence_builder.jlt(iter_reg, 1, batch_exit)
            self.sequence_builder.move(Constants.GRID_TIME, iter_reg, "Rounding up")
            self.sequence_builder.nop()
            self.sequence_builder.label(remainder)
            self.sequence_builder.wait(iter_reg)
            self.sequence_builder.label(batch_exit)

    def ledger(self, duration: Union[int, str], pulse: np.ndarray = None):
        if isinstance(duration, int):
            pulse = np.zeros(duration, dtype=complex) if pulse is None else pulse
            if all((isinstance(p, Number) for p in self._phases)):
                total_phase = sum(self._phases)
                pulse = pulse * np.exp(1.0j * total_phase)
            assert pulse.size == duration
            self._durations.append(duration)
            self._timeline = np.append(self._timeline, pulse)
        elif isinstance(duration, str):
            self._durations.append(duration)
            self._timeline = None  # Destroy the timeline as we are in dynamic setting
        else:
            raise ValueError(f"Expected legal immediate or register, but got {duration}")

    @contextmanager
    def _loop(self, label: str, count: int = 1):
        """
        This context can be used to avoid unrolling loops but needs manual addition of the 0 iteration
        either immediately after the yield or immediately after the context has exited.
        """

        label = self.alloc_mgr.label_gen(label)
        register = self.alloc_mgr.reg_alloc(label)
        self.sequence_builder.move(count, register)
        self.sequence_builder.nop()
        self.sequence_builder.label(label)
        yield register
        self.sequence_builder.loop(register, label)
        self.alloc_mgr.reg_free(register)

    @contextmanager
    def _modulo_reg(self, a: str, n: int):
        """
        Helpful context that prepares a temporary register b such that a ≅ b [n] where
        n is some immediate constant batch size.
        """

        if not isinstance(n, int) or n < 2:
            raise ValueError(
                f"Invalid batch size. Expected an integer >= 2, but got {n} instead"
            )

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")

        iter_reg = self.alloc_mgr.reg_alloc("tmp")
        if a != iter_reg:
            self.sequence_builder.move(
                a,
                iter_reg,
                f"Avoids cluttering {a} as it's likely used as an accumulator",
            )
            self.sequence_builder.nop()

        self.sequence_builder.jlt(iter_reg, n, batch_exit)
        self.sequence_builder.label(batch_enter)
        self.sequence_builder.sub(iter_reg, n, iter_reg)
        self.sequence_builder.nop()
        self.sequence_builder.jge(iter_reg, n, batch_enter)
        self.sequence_builder.label(batch_exit)
        yield iter_reg
        self.alloc_mgr.reg_free(iter_reg)

    def _upd_param_imm(self, duration: int):
        self.sequence_builder.upd_param(Constants.GRID_TIME)
        self.wait_imm(duration - Constants.GRID_TIME)

    def _upd_param_reg(self, duration: str):
        """
        Update latched parameters and wait for `duration` nanoseconds expressed as a register

        If `duration` is less than the threshold Constants.GRID_TIME, the generates assembly rounds up
        and waits for Constants.GRID_TIME.

        Note that the `up_param` instruction is defined only for an immediate operand which must be
        at least Constants.GRID_TIME. Therefore, technically we're already forced to round up.
        """

        batch_enter = self.alloc_mgr.label_gen("batch_enter")
        batch_exit = self.alloc_mgr.label_gen("batch_exit")
        remainder = self.alloc_mgr.label_gen("remainder")

        self.sequence_builder.upd_param(Constants.GRID_TIME)
        with self.alloc_mgr.reg_borrow("tmp") as iter_reg:
            self.sequence_builder.jlt(
                duration,
                Constants.GRID_TIME,
                batch_exit,
                "Guards against risky underflow",
            )
            self.sequence_builder.sub(duration, Constants.GRID_TIME, iter_reg)
            self.sequence_builder.nop()

            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, batch_exit)
            self.sequence_builder.jlt(iter_reg, Constants.MAX_WAIT_TIME, remainder)
            self.sequence_builder.label(batch_enter)
            self.sequence_builder.wait(Constants.MAX_WAIT_TIME)
            self.sequence_builder.sub(iter_reg, Constants.MAX_WAIT_TIME, iter_reg)
            self.sequence_builder.nop()
            self.sequence_builder.jge(iter_reg, Constants.MAX_WAIT_TIME, batch_enter)
            self.sequence_builder.jlt(iter_reg, Constants.GRID_TIME, batch_exit)
            self.sequence_builder.label(remainder)
            self.sequence_builder.wait(iter_reg)
            self.sequence_builder.label(batch_exit)

    def _evaluate_waveform(self, waveform: Waveform, target: PulseChannel):
        """
        The waveform is evaluated as a 1d complex array. In QBlox, the Real and Imag parts of the pulse
        represent the digital offset on the AWG. They both must be within range [-1, 1].
        """

        num_samples = int(calculate_duration(waveform))
        dt = target.sample_time
        length = num_samples * dt
        centre = length / 2.0
        t = np.linspace(
            start=-centre + 0.5 * dt, stop=length - centre - 0.5 * dt, num=num_samples
        )
        pulse = evaluate_shape(waveform, t)
        scale = target.scale
        if isinstance(waveform, (Pulse, CustomPulse)) and waveform.ignore_channel_scale:
            scale = 1

        pulse *= scale
        pulse += target.bias

        if ((pulse.real < -1) | (pulse.real > 1)).any():
            raise ValueError(
                "Voltage range for I exceeded. Make sure all I values are within the range [-1, 1]"
            )
        if ((pulse.imag < -1) | (pulse.imag > 1)).any():
            raise ValueError(
                "Voltage range for Q exceeded. Make sure all I values are within the range [-1, 1]"
            )

        return pulse

    def _evaluate_square_pulse(self, pulse: Pulse, target: PulseChannel):
        bias = target.bias
        scale = 1.0 + 0.0j if pulse.ignore_channel_scale else target.scale

        pulse_amp = pulse.scale_factor * pulse.amp
        pulse_amp = scale * pulse_amp + bias

        if np.abs(pulse_amp.real) > 1 or np.abs(pulse_amp.imag) > 1:
            raise ValueError(
                f"""
                Voltage range for I or Q exceeded. Current values:
                waveform.scale_factor: {pulse.scale_factor}
                waveform.amp: {pulse.amp}
                (Channel) scale: {scale}
                (Channel) bias: {bias}
                """
            )

        return pulse_amp

    def _register_signal(self, waveform, target, data, name):
        index = self.sequence_builder.lookup_waveform_by_data(data)
        if index is not None:
            log.debug(
                f"Reusing signal {name} at index {index} for pulse {waveform} on channel {target}"
            )
        elif data.size > self._wf_memory:
            raise ValueError(
                f"No more waveform memory left for signal {name} of pulse {waveform} on channel {target}"
            )
        else:
            wf_hash = hash(waveform)
            wf_name = type(waveform).__name__
            if isinstance(waveform, Pulse):
                wf_name = waveform.shape.name

            index = self._wf_index
            self.sequence_builder.add_waveform(
                f"{wf_name}_{wf_hash}_{name}", index, data.tolist()
            )
            self._wf_memory = self._wf_memory - data.size
            self._wf_index = self._wf_index + 1

        return index

    def _register_waveform(self, waveform, target, data):
        i_index = self._register_signal(waveform, target, data.real, "I")
        q_index = self._register_signal(waveform, target, data.imag, "Q")
        return i_index, q_index

    def _register_acquisition(self, acq_name: str, num_bins: Union[int, str]):
        acq_index = self._acq_index
        self.sequence_builder.add_acquisition(acq_name, acq_index, num_bins)
        self._acq_index = acq_index + 1
        return acq_index

    def _register_weight(self, name: str, weight: np.ndarray):
        wgt_index = self._wgt_index
        self.sequence_builder.add_weight(name, wgt_index, weight.tolist())
        self._wgt_index = wgt_index + 1
        return wgt_index

    def _do_acquire(self, wgt_name, acq_filter, acq_index, acq_bin, acq_width):
        if acq_filter:
            weight = acq_filter.samples
            wgt_i_index = self._register_weight(f"{wgt_name}_weight_I", weight.real)
            wgt_q_index = self._register_weight(f"{wgt_name}_weight_Q", weight.imag)

            # TODO - silly convoluted way of using the `acquire_weighed` instruction
            # TODO - simplify when Qblox supports more freedom of memory addressing
            with (
                self.alloc_mgr.reg_borrow("wgt_i") as wgt_i_reg,
                self.alloc_mgr.reg_borrow("wgt_q") as wgt_q_reg,
            ):
                self.sequence_builder.move(wgt_i_index, wgt_i_reg)
                self.sequence_builder.move(wgt_q_index, wgt_q_reg)

                self.sequence_builder.acquire_weighed(
                    acq_index, acq_bin, wgt_i_reg, wgt_q_reg, acq_width
                )
        else:
            self.sequence_builder.acquire(acq_index, acq_bin, acq_width)

    @contextmanager
    def _wrapper_pulse(
        self,
        delay_width,
        pulse_width,
        pulse_shape,
        i_steps,
        q_steps,
        i_index,
        q_index,
    ):
        effective_width = max(min(pulse_width, delay_width), Constants.GRID_TIME)
        if pulse_shape == PulseShapeType.SQUARE:
            self.sequence_builder.set_awg_offs(i_steps, q_steps)
            self.sequence_builder.upd_param(effective_width)
        else:
            self.sequence_builder.play(i_index, q_index, effective_width)
        yield pulse_width - effective_width
        if pulse_shape == PulseShapeType.SQUARE:
            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)

            self.ledger(Constants.GRID_TIME)

    @contextmanager
    def _wrapper_cond(self, mask, operator, duration):
        """
        A wrapper for conditional regions
        """

        self.sequence_builder.set_cond(
            1, mask, operator, duration, "Start of conditional region"
        )
        yield
        self.sequence_builder.set_cond(
            0, mask, operator, duration, "End of conditional region"
        )

    def id(self):
        self.sequence_builder.nop()

    def delay(self, inst: Delay):
        if isinstance(inst.duration, Variable):
            duration = self.alloc_mgr.registers[inst.duration.name]
            self.wait_reg(duration)
        elif isinstance(inst.duration, Number):
            duration = int(calculate_duration(inst))
            self.wait_imm(duration)
        else:
            raise ValueError(f"Expected a Variable or a Number but got {inst.duration}")

        self.ledger(duration)

    def shift_frequency(self, inst, target):
        # TODO - discuss w/t quantum_target.fixed_if is True or False
        old_frequency = target.frequency + self._frequency
        new_frequency = old_frequency + inst.frequency
        if new_frequency < target.min_frequency or new_frequency > target.max_frequency:
            raise ValueError(
                f"Cannot shift pulse channel frequency from '{old_frequency}' to '{new_frequency}'"
            )

        self._frequency = self._frequency + inst.frequency
        shifted_nco_freq = target.baseband_if_frequency + self._frequency
        value = QbloxLegalisationPass.freq_as_steps(shifted_nco_freq)  # 1 MHZ <-> 4e6
        self.sequence_builder.set_freq(value)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        self.ledger(Constants.GRID_TIME)

    def shift_phase(self, inst: PhaseShift):
        if isinstance(inst.phase, Variable):
            ph_reg = self.alloc_mgr.registers[inst.phase.name]
            with self._modulo_reg(ph_reg, Constants.NCO_MAX_PHASE_STEPS) as iter_reg:
                self.sequence_builder.set_ph_delta(iter_reg)
            self.sequence_builder.upd_param(Constants.GRID_TIME)
        elif isinstance(inst.phase, Number):
            ph_imm = QbloxLegalisationPass.phase_as_steps(inst.phase)
            self.sequence_builder.set_ph_delta(ph_imm)
            self.sequence_builder.upd_param(Constants.GRID_TIME)
        else:
            raise ValueError(f"Expected a Variable or a Number but got {inst.phase}")

        self._phases.append(inst.phase)
        self.ledger(Constants.GRID_TIME)

    def reset_phase(self):
        self.sequence_builder.reset_ph()
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        self._phases.clear()
        self.ledger(Constants.GRID_TIME)

    def waveform(self, waveform: Waveform, target: PulseChannel):
        attr2var = {
            attr: var for attr, var in vars(waveform).items() if isinstance(var, Variable)
        }

        if isinstance(waveform, Pulse) and waveform.shape == PulseShapeType.SQUARE:
            if "amp" in attr2var:
                pulse_amp = attr2var["amp"]
                i_steps = self.alloc_mgr.registers[pulse_amp.name]
                q_steps = self.alloc_mgr.registers["zero"]
            else:
                pulse_amp = self._evaluate_square_pulse(waveform, target)
                i_steps = int(pulse_amp.real * Constants.MAX_OFFSET)
                q_steps = int(pulse_amp.imag * Constants.MAX_OFFSET)

            self.sequence_builder.set_awg_offs(i_steps, q_steps)

            if "width" in attr2var:
                pulse_width = attr2var["width"]
                pulse_width = self.alloc_mgr.registers[pulse_width.name]
                self._upd_param_reg(pulse_width)
            else:
                pulse_width = int(calculate_duration(waveform))
                if pulse_width < Constants.GRID_TIME:
                    log.debug(
                        f"""
                        Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                        Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                        This pulse will be ignored.
                        """
                    )
                    return
                self._upd_param_imm(pulse_width)

            self.sequence_builder.set_awg_offs(0, 0)
            self.sequence_builder.upd_param(Constants.GRID_TIME)

            self.ledger(Constants.GRID_TIME)

            pulse = (
                np.ones(pulse_width, dtype=complex) * pulse_amp
                if isinstance(pulse_width, Number) and isinstance(pulse_amp, Number)
                else None
            )
        else:
            pulse = self._evaluate_waveform(waveform, target)
            pulse_width = pulse.size
            if pulse_width < Constants.GRID_TIME:
                log.debug(
                    f"""
                    Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                    Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                    This pulse will be ignored.
                    """
                )
                return

            max_width = min(pulse_width, Constants.MAX_WAIT_TIME)
            i_index, q_index = self._register_waveform(waveform, target, pulse)
            self.sequence_builder.play(i_index, q_index, max_width)
            self.wait_imm(pulse_width - max_width)

        self.ledger(pulse_width, pulse)

    @staticmethod
    def synchronize(inst: Synchronize, contexts: Dict):
        """
        Favours static time padding whenever possible, or else uses SYNC.
        TODO - Default to SYNC when Qblox supports finer grained SYNC groups
        """

        is_static = all(
            all((isinstance(d, Number) for d in contexts[target].durations))
            for target in inst.quantum_targets
        )

        for target in inst.quantum_targets:
            cxt = contexts[target]
            if is_static:
                duration_offset = (
                    max((contexts[target].duration for target in inst.quantum_targets))
                    - cxt.duration
                )
                cxt.wait_imm(duration_offset)
                cxt.ledger(duration_offset)
            else:
                cxt.sequence_builder.wait_sync(Constants.GRID_TIME)
                cxt.ledger(Constants.GRID_TIME)

    def device_update(self, du_inst: DeviceUpdate):
        if isinstance(du_inst.value, Variable):
            val_reg = self.alloc_mgr.registers[du_inst.value.name]
            if du_inst.attribute == "frequency":
                self.sequence_builder.set_freq(val_reg)
                self.sequence_builder.upd_param(Constants.GRID_TIME)

                self.ledger(Constants.GRID_TIME)
            elif du_inst.attribute == "scale":
                pass
            else:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {du_inst.attribute} for instruction {du_inst}"
                )
        elif isinstance(du_inst.value, Number):
            if du_inst.attribute == "frequency":
                lo_freq, nco_freq = TILegalisationPass.decompose_freq(
                    du_inst.value, du_inst.target
                )
                freq_imm = QbloxLegalisationPass.freq_as_steps(nco_freq)
                self.sequence_builder.set_freq(freq_imm)
                self.sequence_builder.upd_param(Constants.GRID_TIME)

                self.ledger(Constants.GRID_TIME)
            elif du_inst.attribute == "scale":
                pass
            else:
                raise NotImplementedError(
                    f"Unsupported processing of attribute {du_inst.attribute} for instruction {du_inst}"
                )
        else:
            raise ValueError(f"Expected a Variable or a Number but got {du_inst.value}")


class QbloxContext(AbstractContext):
    def __init__(self, repeat: Repeat):
        super().__init__()

        self._repeat_count = repeat.repeat_count
        self._repeat_period = repeat.repetition_period
        self._repeat_reg = None
        self._repeat_label = None

    def __enter__(self):
        self.clear()
        self._repeat_reg = self.alloc_mgr.reg_alloc("shot")
        self._repeat_label = self.alloc_mgr.label_gen("shot")

        self.sequence_builder.set_mrk(3)
        self.sequence_builder.set_latch_en(1, 4)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        self.sequence_builder.move(0, self._repeat_reg, "Shot / Repeat iteration")

        self.sequence_builder.label(self._repeat_label)
        self.sequence_builder.wait_sync(
            Constants.GRID_TIME, "Sync at the beginning of shot"
        )
        self.reset_phase()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.wait_imm(int(self._repeat_period * 1e9))
        self.sequence_builder.add(
            self._repeat_reg, 1, self._repeat_reg, "Increment Shot / Repeat iterator"
        )
        self.sequence_builder.nop()
        self.sequence_builder.jlt(self._repeat_reg, self._repeat_count, self._repeat_label)
        self.sequence_builder.stop()

        self.alloc_mgr.reg_free(self._repeat_reg)
        self._repeat_label = None
        self._repeat_reg = None

    def measure_acquire(
        self,
        measure: MeasurePulse,
        acquire: Acquire,
        post_procs: List[PostProcessing],
        target: PulseChannel,
    ):
        pulse = self._evaluate_waveform(measure, target)
        pulse_width = pulse.size
        delay_width = int(calculate_duration(Delay(acquire.channel, acquire.delay)))

        if pulse_width < Constants.GRID_TIME:
            log.debug(
                f"""
                Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                This pulse will be ignored. 
                """
            )
            return

        if pulse_width < delay_width + Constants.GRID_TIME:
            raise ValueError(
                f"""
                Expected pulse width >= delay width + {Constants.GRID_TIME} ns.
                Got pulse width = {pulse_width} ns and delay width = {delay_width} ns.
                """
            )

        requires_shot_avg = any(
            (
                pp.process == PostProcessType.MEAN and ProcessAxis.SEQUENCE in pp.axes
                for pp in post_procs
            )
        )
        num_bins = 1 if requires_shot_avg else self._repeat_count
        acq_bin = 0 if requires_shot_avg else self._repeat_reg
        acq_width = int(calculate_duration(acquire))
        acq_index = self._register_acquisition(acquire.output_variable, num_bins)
        self.sequencer_config.square_weight_acq.integration_length = acq_width
        self.sequencer_config.thresholded_acq.rotation = np.rad2deg(acquire.rotation)
        self.sequencer_config.thresholded_acq.threshold = acquire.threshold

        i_steps, q_steps = None, None
        i_index, q_index = None, None
        if measure.shape == PulseShapeType.SQUARE:
            i_steps = int(pulse[0].real * Constants.MAX_OFFSET)
            q_steps = int(pulse[0].imag * Constants.MAX_OFFSET)
        else:
            i_index, q_index = self._register_waveform(measure, target, pulse)

        with self._wrapper_pulse(
            delay_width, pulse_width, measure.shape, i_steps, q_steps, i_index, q_index
        ) as remaining_width:
            assert remaining_width >= Constants.GRID_TIME
            self._do_acquire(
                acquire.output_variable,
                acquire.filter,
                acq_index,
                acq_bin,
                remaining_width,
            )

        self.ledger(pulse_width, pulse)


class NewQbloxContext(AbstractContext):
    def __init__(
        self,
        alloc_mgr: AllocationManager,
        scoping_result: ScopingResult,
        rw_result: ReadWriteResult,
        iter_bounds: Dict[str, IterBound],
    ):
        super().__init__(alloc_mgr)
        self.scope2symbols: Dict[Tuple[Instruction, Optional[Instruction]], Set[str]] = (
            scoping_result.scope2symbols
        )
        self.symbol2scopes: Dict[str, List[Tuple[Instruction, Optional[Instruction]]]] = (
            scoping_result.symbol2scopes
        )
        self.reads: Dict[str, List[Instruction]] = rw_result.reads
        self.writes: Dict[str, List[Instruction]] = rw_result.writes
        self.iter_bounds = iter_bounds

    def __enter__(self):
        """
        Serves as the prologue
        """

        self.sequence_builder.set_mrk(3)
        self.sequence_builder.set_latch_en(1, 4)
        self.sequence_builder.upd_param(Constants.GRID_TIME)

        for name, register in self.alloc_mgr.registers.items():
            self.sequence_builder.move(
                0,
                register,
                f"Precautionary initialisation for variable {name}",
            )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Serves as the epilogue
        """

        self.sequence_builder.stop()

    def measure_acquire(
        self, measure: MeasurePulse, acquire: Acquire, target: PulseChannel
    ):
        pulse = self._evaluate_waveform(measure, target)
        pulse_width = pulse.size
        delay_width = int(calculate_duration(Delay(acquire.channel, acquire.delay)))

        if pulse_width < Constants.GRID_TIME:
            log.debug(
                f"""
                Minimum pulse width is {Constants.GRID_TIME} ns with a resolution of {1} ns.
                Please round up the width to at least {Constants.GRID_TIME} nanoseconds.
                This pulse will be ignored. 
                """
            )
            return

        if pulse_width < delay_width + Constants.GRID_TIME:
            raise ValueError(
                f"""
                Expected pulse width >= delay width + {Constants.GRID_TIME} ns.
                Got pulse width = {pulse_width} ns and delay width = {delay_width} ns.
                """
            )

        name = f"acquire_{hash(acquire)}"
        num_bins = self.iter_bounds[name].count
        acq_bin = self.alloc_mgr.registers[name]
        acq_width = int(calculate_duration(acquire))
        acq_index = self._register_acquisition(acquire.output_variable, num_bins)
        self.sequencer_config.square_weight_acq.integration_length = acq_width
        self.sequencer_config.thresholded_acq.rotation = acquire.rotation
        self.sequencer_config.thresholded_acq.threshold = acquire.threshold

        i_steps, q_steps = None, None
        i_index, q_index = None, None
        if measure.shape == PulseShapeType.SQUARE:
            i_steps = int(pulse[0].real * Constants.MAX_OFFSET)
            q_steps = int(pulse[0].imag * Constants.MAX_OFFSET)
        else:
            i_index, q_index = self._register_waveform(measure, target, pulse)

        with self._wrapper_pulse(
            delay_width, pulse_width, measure.shape, i_steps, q_steps, i_index, q_index
        ) as remaining_width:
            assert remaining_width >= Constants.GRID_TIME
            self._do_acquire(
                acquire.output_variable,
                acquire.filter,
                acq_index,
                acq_bin,
                remaining_width,
            )

        self.ledger(pulse_width, pulse)

    @staticmethod
    def enter_repeat(inst: Repeat, contexts: Dict):
        iter_name = f"repeat_{hash(inst)}"
        for context in contexts.values():
            register = context.alloc_mgr.registers[iter_name]
            bound = context.iter_bounds[iter_name]
            context.sequence_builder.move(bound.start, register, "Shot / Repeat iteration")

            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.label(label)
            context.sequence_builder.wait_sync(
                Constants.GRID_TIME, "Sync at the beginning of shot"
            )
            context.reset_phase()

    @staticmethod
    def exit_repeat(inst: Repeat, contexts: Dict):
        iter_name = f"repeat_{hash(inst)}"
        for context in contexts.values():
            register = context.alloc_mgr.registers[iter_name]
            label = context.alloc_mgr.labels[iter_name]
            bound = context.iter_bounds[iter_name]

            context.wait_imm(int(inst.repetition_period * 1e9))
            context.sequence_builder.add(
                register, bound.step, register, "Increment Shot / Repeat iterator"
            )
            context.sequence_builder.nop()
            context.sequence_builder.jlt(register, bound.end + bound.step, label)

    @staticmethod
    def enter_sweep(inst: Sweep, contexts: Dict):
        iter_name = f"sweep_{hash(inst)}"
        for context in contexts.values():
            var_names = [n for n in inst.variables if n in context.reads] + [iter_name]
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.move(
                    bound.start,
                    register,
                    f"Initialise register allocated for variable {name}",
                )

            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.label(label)

    @staticmethod
    def exit_sweep(inst: Sweep, contexts: Dict):
        iter_name = f"sweep_{hash(inst)}"
        for context in contexts.values():
            var_names = [n for (n, insts) in context.writes.items() if inst in insts]
            for name in var_names:
                register = context.alloc_mgr.registers[name]
                bound = context.iter_bounds[name]
                context.sequence_builder.add(
                    register,
                    bound.step,
                    register,
                    f"Increment register for variable {name}",
                )
                context.sequence_builder.nop()

            register = context.alloc_mgr.registers[iter_name]
            bound = context.iter_bounds[iter_name]
            label = context.alloc_mgr.labels[iter_name]
            context.sequence_builder.jlt(register, bound.end + bound.step, label)


class QbloxCFGWalker(DfsTraversal):
    def __init__(self, contexts: Dict[PulseChannel, NewQbloxContext]):
        super().__init__()
        self.contexts = contexts

    def enter(self, block):
        iterator = block.iterator()
        while (inst := next(iterator, None)) is not None:
            if isinstance(inst, Sweep):
                NewQbloxContext.enter_sweep(inst, self.contexts)
            elif isinstance(inst, Repeat):
                NewQbloxContext.enter_repeat(inst, self.contexts)
            elif isinstance(inst, QuantumInstruction):
                if isinstance(inst, PostProcessing):
                    continue
                elif isinstance(inst, Synchronize):
                    NewQbloxContext.synchronize(inst, self.contexts)
                    continue

                for target in inst.quantum_targets:
                    context = self.contexts[target]
                    if isinstance(inst, DeviceUpdate):
                        context.device_update(inst)
                    elif isinstance(inst, PhaseReset):
                        context.reset_phase()
                    elif isinstance(inst, MeasurePulse):
                        next_inst = next(iterator, None)
                        if next_inst is None or not isinstance(next_inst, Acquire):
                            raise ValueError(
                                "Found a MeasurePulse but no Acquire instruction followed"
                            )

                        context.measure_acquire(inst, next_inst, target)
                    elif isinstance(inst, Waveform):
                        context.waveform(inst, target)
                    elif isinstance(inst, Delay):
                        context.delay(inst)
                    elif isinstance(inst, PhaseShift):
                        context.shift_phase(inst)
                    elif isinstance(inst, FrequencyShift):
                        context.shift_frequency(inst, target)
                    elif isinstance(inst, Id):
                        context.id()

    def exit(self, block):
        iterator = block.iterator()
        while (inst := next(iterator, None)) is not None:
            if isinstance(inst, Repeat):
                NewQbloxContext.exit_repeat(inst, self.contexts)
            elif isinstance(inst, Sweep):
                NewQbloxContext.exit_sweep(inst, self.contexts)


class QbloxBackend1(BaseBackend, InvokerMixin):
    def build_pass_pipeline(self, *args, **kwargs):
        pass

    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: Optional[ResultManager] = None,
        met_mgr: Optional[MetricsManager] = None,
        compiler_config: Optional[CompilerConfig] = None,
    ) -> Dict[int, QbloxExecutable]:
        triage_result: TriageResult = res_mgr.lookup_by_type(TriageResult)
        sweeps = triage_result.sweeps

        switerator = SweepIterator.from_sweeps(sweeps)
        dinjectors = DeviceInjectors(triage_result.device_updates)
        ordered_executables: Dict[int, QbloxExecutable] = OrderedDict()
        try:
            dinjectors.inject()
            while not switerator.is_finished():
                switerator.do_sweep(triage_result.quantum_instructions)
                ordered_executables[switerator.accumulated_sweep_iteration] = self._do_emit(
                    triage_result
                )
            return ordered_executables
        except BaseException as e:
            raise e
        finally:
            switerator.revert(triage_result.quantum_instructions)
            dinjectors.revert()

    def _do_emit(self, triage_result: TriageResult) -> QbloxExecutable:
        repeat = next(iter(triage_result.repeats))
        contexts: Dict[PulseChannel, QbloxContext] = {}

        with ExitStack() as stack:
            inst_iter = iter(triage_result.quantum_instructions)
            while (inst := next(inst_iter, None)) is not None:
                if isinstance(inst, PostProcessing):
                    continue  # Ignore postprocessing

                if isinstance(inst, Synchronize):
                    for target in inst.quantum_targets:
                        if not isinstance(target, PulseChannel):
                            raise ValueError(f"{target} is not a PulseChannel")
                        contexts.setdefault(
                            target, stack.enter_context(QbloxContext(repeat))
                        )
                    QbloxContext.synchronize(inst, contexts)
                    continue

                for target in inst.quantum_targets:
                    if not isinstance(target, PulseChannel):
                        raise ValueError(f"{target} is not a PulseChannel")

                    context = contexts.setdefault(
                        target, stack.enter_context(QbloxContext(repeat))
                    )

                    if isinstance(inst, DeviceUpdate):
                        context.device_update(inst)
                    elif isinstance(inst, PhaseReset):
                        context.reset_phase()
                    elif isinstance(inst, MeasurePulse):
                        acquire = next(inst_iter, None)
                        if acquire is None or not isinstance(acquire, Acquire):
                            raise ValueError(
                                "Found a MeasurePulse but no Acquire instruction followed"
                            )
                        post_procs = triage_result.pp_map[acquire.output_variable]
                        context.measure_acquire(inst, acquire, post_procs, target)
                    elif isinstance(inst, Waveform):
                        context.waveform(inst, target)
                    elif isinstance(inst, Delay):
                        context.delay(inst)
                    elif isinstance(inst, PhaseShift):
                        context.shift_phase(inst)
                    elif isinstance(inst, FrequencyShift):
                        context.shift_frequency(inst, target)
                    elif isinstance(inst, Id):
                        context.id()

        packages = {
            target: context.create_package()
            for target, context in contexts.items()
            if not context.is_empty()
        }

        return QbloxExecutable(packages=packages, triage_result=triage_result)


class QbloxBackend2(BaseBackend, InvokerMixin):
    def build_pass_pipeline(self, *args, **kwargs):
        return PassManager() | PreCodegenPass() | CFGPass()

    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: Optional[ResultManager] = None,
        met_mgr: Optional[MetricsManager] = None,
        compiler_config: Optional[CompilerConfig] = None,
    ) -> Executable:
        self.run_pass_pipeline(ir, res_mgr, met_mgr)

        triage_result: TriageResult = res_mgr.lookup_by_type(TriageResult)
        binding_result: BindingResult = res_mgr.lookup_by_type(BindingResult)
        precodegen_result: PreCodegenResult = res_mgr.lookup_by_type(PreCodegenResult)
        cfg_result: CFGResult = res_mgr.lookup_by_type(CFGResult)

        scoping_results: Dict[PulseChannel, ScopingResult] = binding_result.scoping_results
        rw_results: Dict[PulseChannel, ReadWriteResult] = binding_result.rw_results
        iter_bound_results: Dict[PulseChannel, Dict[str, IterBound]] = (
            binding_result.iter_bound_results
        )
        alloc_mgrs: Dict[PulseChannel, AllocationManager] = precodegen_result.alloc_mgrs

        with ExitStack() as stack:
            contexts = {
                t: stack.enter_context(
                    NewQbloxContext(
                        alloc_mgr=alloc_mgrs[t],
                        scoping_result=scoping_results[t],
                        rw_result=rw_results[t],
                        iter_bounds=iter_bound_results[t],
                    )
                )
                for t in triage_result.target_map
            }
            QbloxCFGWalker(contexts).run(cfg_result.cfg)

        packages = {
            target: context.create_package()
            for target, context in contexts.items()
            if not context.is_empty()
        }

        return QbloxExecutable(packages=packages, triage_result=triage_result)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from abc import abstractmethod

from compiler_config.config import InlineResultsProcessing
from pydantic import BaseModel, NonNegativeInt, PositiveInt
from pydantic_core import from_json

from qat.ir.instructions import Assign
from qat.ir.measure import PostProcessing
from qat.purr.compiler.instructions import AcquireMode
from qat.utils.pydantic import RehydratableModel


class AcquireData(BaseModel):
    """Contains acquisition information for a measurement / readout that is needed at
    runtime.

    :param NonNegativeInt length: A readout is performed for some given time which
        translates to a number of discrete samples.
    :param NonNegativeInt position: The sample which a readout starts.
    :param AcquireMode mode: The acqusition mode used by the hardware to carry out the
        readout.
    :param str output_variable: The name of the variable to save the result to.
    """

    length: NonNegativeInt
    position: NonNegativeInt
    mode: AcquireMode
    output_variable: str


class Executable(RehydratableModel):
    """
    :class:`Executables` are instruction packages that will be sent to a target (either
    physical control hardware or a simulator).

    This class serves as a base model that can be inherited to define a complete instruction
    package tailored to a particular target.

    :param channel_data: Stores the data required by the control hardware for each pulse
        channel.
    :type channel_data: dict[str, ChannelData]
    :param PositiveInt shots: The number of times the program is executed.
    :param PositiveInt compiled_shots: When the required number of shots exceeds the
        allowed amount by the target machine, shots can be batched into groups. This states
        how many shots to do in each batch.
    :param post_processing: Contains the post-processing information for each acquisition.
    :type post_processing: dict[str, List[PostProcessing]]
    :param results_processing: Contains the information for how results should be formatted.
    :type results_processing: dict[str, InlineResultsProcessing]
    :param assigns: Assigns results to given variables.
    :type assigns: list[Assign]
    :param returns: Which acqusitions/variables should be returned.
    :type returns: list[str]
    :param calibration_id: The (unique) calibration ID of the underlying hardware model.
    :type calibration_id: str
    """

    shots: PositiveInt = 1000
    compiled_shots: NonNegativeInt = 0
    post_processing: dict[str, list[PostProcessing]] = dict()
    results_processing: dict[str, InlineResultsProcessing] = dict()
    assigns: list[Assign] = []
    returns: set[str] = set()

    calibration_id: str = ""

    def serialize(self, indent: int = 4) -> str:
        """Serializes the executable as a JSON blob."""
        return self.model_dump_json(indent=indent, exclude_none=True)

    @classmethod
    def deserialize(cls, blob: str):
        """Instantiates a executable from a JSON blob."""
        return cls._rehydrate_object(from_json(blob))

    @property
    @abstractmethod
    def acquires(self) -> list[AcquireData]:
        """Retrieves all acquires for the program.

        This abstract property should be specified for an executable that is coupled to a
        specific backend and target device. It is expected that this returns a list of all
        acquisitions for the program, which will be required by the Runtime."""
        ...


class ChannelData(BaseModel):
    """Base class for "channel data". This can be used for exectuables which target devices
    that seperate instructions over multiple "channels".

    It is expected to contain any target-specific information, along with a property that
    retrieves all acquistion data for that channel.
    """

    acquires: list[AcquireData] | AcquireData = []


class ChannelExecutable(Executable):
    """A base Executable object used for compiled programs that target hardware with many
    "channels", e.g., live hardware. Defines the acquire property to fetch acquisitions from
    the channel data."""

    channel_data: dict[str, ChannelData] = {}

    @property
    def acquires(self) -> list[AcquireData]:
        """Retrieves all acquires from each channel."""

        acquires = []
        for channel in self.channel_data.values():
            if isinstance(channel.acquires, list):
                acquires.extend(channel.acquires)
            else:
                acquires.append(channel.acquires)
        return acquires



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import hashlib
import itertools as it
import random
from collections import defaultdict
from pathlib import Path

import networkx as nx
import numpy as np

from qat.model.builder import PhysicalHardwareModelBuilder
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalBaseband,
    PhysicalChannel,
    Qubit,
    QubitCoupling,
    Resonator,
)
from qat.purr.compiler.hardware_models import QuantumHardwareModel


def random_connectivity(n, max_degree=3, seed=42):
    """
    Generates a random undirected graph but enforcing that the resulting graph is connected.
    """
    edges = list(it.combinations(range(n), 2))
    random.Random(seed).shuffle(edges)
    G = nx.Graph()
    G.add_nodes_from(range(n))
    for node_edges in edges:
        if (
            len(G.edges(node_edges[0])) < max_degree
            and len(G.edges(node_edges[1])) < max_degree
        ):
            G.add_edge(*node_edges)

    return {node: set(neighbors) for node, neighbors in G.adjacency()}


def random_directed_connectivity(n, max_degree=3, seed=42):
    """
    Generates a random directed graph but enforcing that the resulting graph is connected.
    """
    edges = list(it.combinations(range(n), 2))
    random.Random(seed).shuffle(edges)
    G = nx.DiGraph()
    G.add_nodes_from(range(n))
    for node_edges in edges:
        if (
            len(G.edges(node_edges[0])) < max_degree
            and len(G.edges(node_edges[1])) < max_degree
        ):
            G.add_edge(*node_edges)

    return {node: set(neighbors) for node, neighbors in G.adjacency()}


def random_quality_map(connectivity, seed=42):
    coupling_map = {}
    for q1_index, connected_qubits in connectivity.items():
        for q2_index in connected_qubits:
            coupling_map[(q1_index, q2_index)] = random.Random(seed).uniform(0.0, 1.0)
    return coupling_map


def pick_subconnectivity(connectivity, n, seed=42):
    sub_qubits = random.Random(seed).sample(list(connectivity.keys()), n)
    sub_connectivity = defaultdict(set)
    for qubit in sub_qubits:
        for connected_qubit in connectivity[qubit]:
            if qubit not in sub_connectivity[connected_qubit]:
                sub_connectivity[qubit].add(connected_qubit)

    return sub_connectivity


def generate_connectivity_data(n_qubits, n_logical_qubits, seed=42):
    physical_connectivity = random_connectivity(n=n_qubits, seed=seed)
    logical_connectivity = pick_subconnectivity(
        physical_connectivity, n=n_logical_qubits, seed=seed
    )
    logical_connectivity_quality = random_quality_map(
        connectivity=logical_connectivity, seed=seed
    )
    return (physical_connectivity, logical_connectivity, logical_connectivity_quality)


def generate_hw_model(n_qubits, seed=42):
    physical_connectivity, _, _ = generate_connectivity_data(n_qubits, n_qubits, seed=seed)

    builder = PhysicalHardwareModelBuilder(physical_connectivity=physical_connectivity)
    return builder.model


def apply_setup_to_echo_hardware(qubit_count: int, connectivity) -> QuantumHardwareModel:
    qubit_devices = []
    resonator_devices = []
    channel_index = 1

    qubit_devices = []
    resonator_devices = []
    channel_index = 1
    hw = QuantumHardwareModel()
    for primary_index in range(qubit_count):
        bb1 = PhysicalBaseband(f"LO{channel_index}", 5.5e9)
        bb2 = PhysicalBaseband(f"LO{channel_index + 1}", 8.5e9)
        hw.add_physical_baseband(bb1, bb2)

        ch1 = PhysicalChannel(f"CH{channel_index}", 1.0e-9, bb1, 1)
        ch2 = PhysicalChannel(
            f"CH{channel_index + 1}", 1.0e-9, bb2, 1, acquire_allowed=True
        )
        hw.add_physical_channel(ch1, ch2)

        resonator = Resonator(f"R{primary_index}", ch2)
        resonator.create_pulse_channel(ChannelType.measure, frequency=8.5e9)
        resonator.create_pulse_channel(ChannelType.acquire, frequency=8.5e9)

        qubit = Qubit(primary_index, resonator, ch1)
        qubit.create_pulse_channel(ChannelType.drive, frequency=5.5e9)

        qubit_devices.append(qubit)
        resonator_devices.append(resonator)
        channel_index = channel_index + 2

    qubits_by_index = {qb.index: qb for qb in qubit_devices}
    for connection in connectivity:
        left_index, right_index = connection
        qubit_left = qubits_by_index.get(left_index)
        qubit_right = qubits_by_index.get(right_index)

        qubit_left.create_pulse_channel(
            auxiliary_devices=[qubit_right],
            channel_type=ChannelType.cross_resonance,
            frequency=5.5e9,
            scale=50,
        )
        qubit_left.create_pulse_channel(
            auxiliary_devices=[qubit_right],
            channel_type=ChannelType.cross_resonance_cancellation,
            frequency=5.5e9,
            scale=0.0,
        )
        qubit_left.add_coupled_qubit(qubit_right)
        hw.qubit_direction_couplings.append(
            QubitCoupling(direction=connection, quality=random.Random().uniform(0.0, 1.0))
        )

    hw.add_quantum_device(*qubit_devices, *resonator_devices)
    hw.is_calibrated = True
    return hw


def generate_random_linear(qubit_indices):
    output = {}
    for index in qubit_indices:
        random_0 = random.random()
        random_1 = random.random()
        output[index] = np.array([[random_0, 1 - random_1], [1 - random_0, random_1]])
    return output


def check_type_legacy_or_pydantic(hw_model: QuantumHardwareModel | PydHardwareModel):
    if not isinstance(hw_model, QuantumHardwareModel | PydHardwareModel | None):
        raise TypeError(
            f"Invalid type for the hardware model: {hw_model.__class__.__name__}. Please provide a `QuantumHardwareModel` or `PhysicalHardwareModel`."
        )

    return hw_model


def hash_calibration_file(file_path: str | Path, algorithm="md5", chunk_size=8_192) -> str:
    """Compute the hash of a calibration file."""
    file_path = Path(file_path)

    hash_func = hashlib.new(algorithm)
    with file_path.open("rb") as file:
        # Byte chunks from the file and update hash.
        while chunk := file.read(chunk_size):
            hash_func.update(chunk)

    return hash_func.hexdigest()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import abc
import warnings
from typing import Dict, List, Optional

import numpy as np

from qat.core.config.configure import get_config
from qat.core.pass_base import PassManager
from qat.engines import ConnectionMixin, NativeEngine
from qat.executables import AcquireData
from qat.purr.compiler.instructions import AcquireMode
from qat.runtime.connection import ConnectionMode
from qat.runtime.passes.transform import (
    AssignResultsTransform,
    InlineResultsProcessingTransform,
    PostProcessingTransform,
)


class BaseRuntime(abc.ABC):
    """Provides a Base class to build on for runtimes of varying complexities.

    A runtime provides the means to execute quantum programs. It can take on various
    responsibilities, including interfacing the execution engine and post-processing of
    results. Runtimes are designed to fit a specific purpose. For example, the
    :class:`SimpleRuntime` provides the means to execute already compiled programs
    :class:`Executable`. In the future, there will be support for hybrid runtimes that take
    on both compilation and execution responsibilities.
    """

    def __init__(
        self,
        engine: NativeEngine,
        results_pipeline: Optional[PassManager] = None,
        connection_mode: ConnectionMode = ConnectionMode.DEFAULT,
    ):
        """
        :param engine: The execution engine for a target machine.
        :param results_pipeline: Optionally provided a pipeline for results processing. If
            not provided, a default pipeline is provided.
        :param connection_mode: Specifies how the connection is maintained.
        """
        self.engine = engine
        self.connection_mode = connection_mode
        self.connect_engine(ConnectionMode.CONNECT_AT_BEGINNING)

        if not results_pipeline:
            results_pipeline = self.default_pipeline()
        self.results_pipeline = results_pipeline

    @abc.abstractmethod
    def execute(package, *args): ...

    def default_pipeline(self):
        return (
            PassManager()
            | PostProcessingTransform()
            | InlineResultsProcessingTransform()
            | AssignResultsTransform()
        )

    @staticmethod
    def number_of_batches(total_shots: int, shots_per_batch: int):
        """Calculates the number of shot batches to execute.

        When the total number of shots exceeds the capabilities of the target machine, we can
        execute a number of batches with a subset of the shots. This number of shots should
        be calculated during compilation, and included in the executable. This method
        calculates number of batches to execute.

        In the instance that the total number of shots cannot be batched into a whole number
        of the compiled shots, the runtime will execute more shots than required. If results
        are returned from the hardware per shot, then we can simply trim the results down to
        the required amount of shots. However, this cannot be done if the post-processing
        over the shots is done on the hardware. In this case, the program will be executed
        for :code:`ceil(total_shots / shots_per_batch) * shots_per_batch` shots.

        If the compiled number of shots is zero, then it is assumed all shots can be
        achieved in a single batch.

        :param int total_shots: The total number of shots to execute.
        :param int shots_per_batch: The compiled number of shots that can be executed in a
            single batch.
        """

        if not isinstance(total_shots, int) and total_shots < 0:
            raise ValueError("The number of shots must be a non-negative integer.")

        if not isinstance(shots_per_batch, int) and (shots_per_batch < 0):
            raise ValueError(
                "The shots per batch must be a positive integer, or `0` to indicate to use "
                "the `total_shots`."
            )

        if shots_per_batch == 0:
            return 1

        number_of_batches = int(np.ceil(total_shots / shots_per_batch))

        if total_shots % shots_per_batch != 0:
            warnings.warn(
                f"Cannot batch {total_shots} into whole batches of {shots_per_batch}."
                "SCOPE acquisitions will be done with a total of "
                f"{shots_per_batch * number_of_batches}."
            )
        return number_of_batches

    @staticmethod
    def validate_max_shots(shots: int):
        qatconfig = get_config()

        # TODO: determine if this should be a pass.
        if shots > qatconfig.MAX_REPEATS_LIMIT:
            raise ValueError(
                f"Number of shots {shots} exceeds the maximum amount of "
                f"{qatconfig.MAX_REPEATS_LIMIT}."
            )

    def connect_engine(self, flag: ConnectionMode) -> bool | None:
        """Connect the engine according to the connection mode."""
        if not isinstance(self.engine, ConnectionMixin):
            return None

        if flag in self.connection_mode:
            if not self.engine.is_connected:
                self.engine.connect()
        return self.engine.is_connected

    def disconnect_engine(self, flag: ConnectionMode) -> bool | None:
        """Disconnect the engine according to the connection mode."""
        if not isinstance(self.engine, ConnectionMixin):
            return None

        if flag in self.connection_mode:
            if self.engine.is_connected:
                self.engine.disconnect()
        return self.engine.is_connected

    def __del__(self):
        self.disconnect_engine(ConnectionMode.DISCONNECT_AT_END)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.disconnect_engine(ConnectionMode.DISCONNECT_AT_END)


class ResultsAggregator:
    """Aggregates the acquisition results from batching of shots."""

    def __init__(self, acquires: List[AcquireData]):
        """Begin the aggregation of results for a list of acquisitions.

        :param acquires: List of acquires to be collected.
        :type acquires: List[AcquireData]
        """
        self._results = {}
        for acquire in acquires:
            match acquire.mode:
                case AcquireMode.RAW:
                    result = np.zeros((0, acquire.length), dtype=complex)
                case AcquireMode.INTEGRATOR:
                    result = np.zeros(0, dtype=complex)
                case AcquireMode.SCOPE:
                    result = np.zeros(acquire.length, dtype=complex)

            self._results[acquire.output_variable] = {
                "mode": acquire.mode,
                "result": result,
                "batches": 0,
            }

    def update(self, new_results: Dict[str, np.ndarray]):
        """Add a batch of results.

        For :attr:`AcquireMode.RAW` and :attr:`AcquireMode.INTEGRATOR`, this means
        to append the new results. For :attr:`AcquireMode.SCOPE`, results are accumulated
        as a sum.
        """
        for output_variable, new_result in new_results.items():
            result = self._results[output_variable]
            result["batches"] += 1
            match result["mode"]:
                case AcquireMode.RAW:
                    result["result"] = np.concatenate(
                        (result["result"], new_result), axis=0
                    )
                case AcquireMode.INTEGRATOR:
                    result["result"] = np.concatenate((result["result"], new_result))
                case AcquireMode.SCOPE:
                    result["result"] += new_result

    def results(self, max_shots: Optional[int] = None):
        """Returns the results normalized in an appropiate way.

        For :attr:`AcquireMode.SCOPE`, results are averaged over batches. For
        :attr:`AcquireMode.RAW` or :attr:`AcquireMode.INTEGRATOR`, a truncated
        amount of :code:`max_shots` shots are returned. If :code:`None`, then returns all
        shots.
        """

        normalized_results = {}
        for output_variable, result in self._results.items():
            match result["mode"]:
                case AcquireMode.RAW:
                    normalized_result = result["result"][0:max_shots, :]
                case AcquireMode.INTEGRATOR:
                    normalized_result = result["result"][0:max_shots]
                case AcquireMode.SCOPE:
                    normalized_result = result["result"] / result["batches"]
            normalized_results[output_variable] = normalized_result
        return normalized_results



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.convert_purr import convert_purr_echo_hw_to_pydantic
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.loaders.base import BasePhysicalModelLoader
from qat.purr.backends.echo import Connectivity

from .purr.echo import EchoModelLoader as LegacyEchoModelLoader


class EchoModelLoader(BasePhysicalModelLoader):
    def __init__(
        self,
        qubit_count: int = 4,
        connectivity: Connectivity | list[(int, int)] | None = None,
    ):
        self._legacy = LegacyEchoModelLoader(
            qubit_count=qubit_count, connectivity=connectivity
        )

    def load(self) -> PhysicalHardwareModel:
        legacy_model = self._legacy.load()
        return convert_purr_echo_hw_to_pydantic(legacy_model)


PydEchoModelLoader = EchoModelLoader



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.backend.waveform_v1.codegen import PydWaveformV1Backend
from qat.frontend import AutoFrontend
from qat.middleend.middleends import ExperimentalDefaultMiddleend
from qat.model.convert_purr import convert_purr_echo_hw_to_pydantic
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import CompilePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class WaveformV1CompilePipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the :class:`PydWaveformV1Backend`.

    .. warning::

        This pipeline is for compilation purposes only and does not execute programs.
        Please select an appropriate execution pipeline if you wish to execute compiled
        programs.

        This pipeline is experimental and still in progress. Please use with caution.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None,
        engine: None = None,
    ) -> CompilePipeline:
        """Constructs a pipeline equipped with the :class:`PydWaveformV1Backend`
        and :class:`EchoEngine`."""

        if engine is not None:
            log.warning(
                "The compilation pipeline does not require an engine. It will be ignored."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        pyd_model = convert_purr_echo_hw_to_pydantic(model)
        return CompilePipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontend(model),
            middleend=ExperimentalDefaultMiddleend(model, pyd_model, target_data),
            backend=PydWaveformV1Backend(pyd_model, model, target_data),
            name=config.name,
        )



from qat.backend.waveform_v1.codegen import WaveformV1Backend as WaveformV1Backend
from qat.backend.waveform_v1.executable import (
    WaveformV1ChannelData as WaveformV1ChannelData,
)
from qat.backend.waveform_v1.executable import WaveformV1Executable as WaveformV1Executable



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from .middleends import CustomMiddleend as CustomMiddleend
from .middleends import DefaultMiddleend as DefaultMiddleend
from .middleends import ExperimentalDefaultMiddleend as ExperimentalDefaultMiddleend
from .middleends import FallthroughMiddleend as FallthroughMiddleend



from collections import defaultdict
from dataclasses import dataclass, field

from qat.ir.instructions import Assign, Instruction, ResultsProcessing, Return
from qat.ir.measure import Acquire, PostProcessing
from qat.model.device import PulseChannel


@dataclass
class PartitionedIR:
    """Stores the results of the :class:`PartitionByPulseChannel`."""

    # TODO: When refactoring into Pydantic instructions, we should replace this object with
    # a Pydantic base model. Ideally, we would unify the different IR levels by making a
    # general "QatIR` flexible enough to support them all, but that might be too optimistic. (COMPILER-382)

    # TODO: Remove `PulseChannel` as keys once we fully migrated to pydantic passes 581
    target_map: dict[PulseChannel | str, list[Instruction]] = field(
        default_factory=lambda: defaultdict(list)
    )
    # TODO: Remove Repeat type option: COMPILER-451
    shots: int | None = field(default_factory=lambda: None)
    compiled_shots: int | None = field(default=None)
    returns: list[Return] = field(default_factory=list)
    assigns: list[Assign] = field(default_factory=list)
    acquire_map: dict[PulseChannel | str, list[Acquire]] = field(
        default_factory=lambda: defaultdict(list)
    )
    pp_map: dict[str, list[PostProcessing]] = field(
        default_factory=lambda: defaultdict(list)
    )
    rp_map: dict[str, ResultsProcessing] = field(default_factory=dict)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import abc

from qat.model.hardware_model import PydLogicalHardwareModel, PydPhysicalHardwareModel
from qat.purr.compiler.hardware_models import LegacyHardwareModel


class BaseModelLoader(abc.ABC):
    """ModelLoaders load HardwareModels from a source configured on initialisation."""

    def __init__(self): ...

    @abc.abstractmethod
    def load(self) -> LegacyHardwareModel | PydLogicalHardwareModel:
        """Load and return the Hardware Model.

        :return: A loaded Hardware Model
        :rtype: LegacyHardwareModel | LogicalHardwareModel
        """
        pass


class BaseLogicalModelLoader(BaseModelLoader):
    @abc.abstractmethod
    def load(self) -> PydPhysicalHardwareModel:
        pass


class BasePhysicalModelLoader(BaseModelLoader):
    @abc.abstractmethod
    def load(self) -> PydPhysicalHardwareModel:
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.backend.waveform_v1.codegen import PydWaveformV1Backend
from qat.engines.waveform_v1 import EchoEngine
from qat.frontend import AutoFrontendWithFlattenedIR
from qat.middleend.middleends import ExperimentalDefaultMiddleend
from qat.model.convert_purr import convert_purr_echo_hw_to_pydantic
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import Pipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime import SimpleRuntime
from qat.runtime.results_pipeline import (
    get_experimental_results_pipeline,
)

log = get_default_logger()


class EchoPipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the :class:`PydWaveformV1Backend`
    and executes them using the :class:`EchoEngine`.

    An engine cannot be provided to the pipeline, as the EchoEngine is used directly.

        .. warning::

        This pipeline is experimental and still in progress. Please use with caution.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None,
        engine: None = None,
    ) -> Pipeline:
        """Constructs a pipeline equipped with the :class:`PydWaveformV1Backend`
        and :class:`EchoEngine`."""

        if engine is not None:
            log.warning(
                "An engine was provided to the EchoPipeline, but it will be ignored. "
                "The EchoEngine is used directly."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        pyd_model = convert_purr_echo_hw_to_pydantic(model)
        results_pipeline = get_experimental_results_pipeline(model, pyd_model)
        return Pipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontendWithFlattenedIR(model),
            middleend=ExperimentalDefaultMiddleend(model, pyd_model, target_data),
            backend=PydWaveformV1Backend(pyd_model, model, target_data),
            runtime=SimpleRuntime(engine=EchoEngine(), results_pipeline=results_pipeline),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

import uuid
from copy import deepcopy
from typing import Optional

from pydantic import Field, field_validator, model_validator
from pydantic_extra_types.semantic_version import SemanticVersion
from semver import Version

from qat.model.device import Qubit, Resonator
from qat.model.error_mitigation import ErrorMitigation
from qat.utils.pydantic import (
    CalibratableUnitInterval,
    FrozenDict,
    FrozenSet,
    NoExtraFieldsModel,
    QubitCoupling,
    QubitId,
)

VERSION = Version(0, 0, 1)


class LogicalHardwareModel(NoExtraFieldsModel):
    """Models a hardware with a given connectivity.

    :param version: Semantic version of the hardware model.
    :param logical_connectivity: Connectivity of the qubits in the hardware model (directed graph).
    """

    version: SemanticVersion = Field(frozen=True, repr=False, default=VERSION)
    logical_connectivity: FrozenDict[QubitId, FrozenSet[QubitId]]

    calibration_id: str = Field(default_factory=lambda: str(uuid.uuid4()), frozen=True)

    @field_validator("version")
    def version_compatibility(version: Version):
        assert version.major == VERSION.major, (
            f"Direct instantiation requires major version compatibility (expected {VERSION.major}.Y.Z, found {version})"
        )
        assert VERSION >= version, (
            f"Latest supported hardware model version {VERSION}, found {version}"
        )
        return VERSION

    def __eq__(self, other: LogicalHardwareModel) -> bool:
        if type(self) is not type(other):
            return False

        if self.__class__.model_fields != other.__class__.model_fields:
            return False

        if self.version != other.version:
            return False

        if self.logical_connectivity != other.logical_connectivity:
            return False

        return True

    def __ne__(self, other: LogicalHardwareModel) -> bool:
        return not self.__eq__(other)


class PhysicalHardwareModel(LogicalHardwareModel):
    """Class for calibrating our QPU hardware.

    :param qubits: The superconducting qubits on the chip.
    :param physical_connectivity: The connectivities of the physical qubits on the QPU (undirected graph).
    :param logical_connectivity: The connectivities (directed graph) of the qubits used for compilation, which can be a subgraph of `physical_connectivity`.
    :param logical_connectivity_quality: Quality of the connections between the qubits.
    :param error_mitigation: Error mitigation strategy for this hardware model.
    """

    qubits: FrozenDict[QubitId, Qubit]
    physical_connectivity: FrozenDict[QubitId, FrozenSet[QubitId]] = Field(frozen=True)
    logical_connectivity: Optional[FrozenDict[QubitId, FrozenSet[QubitId]]] = Field(
        default=None
    )
    logical_connectivity_quality: FrozenDict[QubitCoupling, CalibratableUnitInterval]
    error_mitigation: ErrorMitigation = ErrorMitigation()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._generate_mappings()

    def _generate_mappings(self):
        self._ids_to_pulse_channels = {}
        self._ids_to_physical_channels = {}
        self._pulse_channel_ids_to_physical_channel = {}
        self._pulse_channel_ids_to_device = {}
        self._qubits_to_qubit_ids = {}
        self._resonators_to_qubits = {}

        for qubit_id, qubit in self.qubits.items():
            self._qubits_to_qubit_ids[qubit] = qubit_id
            self._resonators_to_qubits[qubit.resonator] = qubit

            for device in [qubit, qubit.resonator]:
                phys_channel = device.physical_channel
                self._ids_to_physical_channels[phys_channel.uuid] = phys_channel

                for pulse_channel in device.all_pulse_channels:
                    self._ids_to_pulse_channels[pulse_channel.uuid] = pulse_channel
                    self._pulse_channel_ids_to_physical_channel[pulse_channel.uuid] = (
                        phys_channel
                    )
                    self._pulse_channel_ids_to_device[pulse_channel.uuid] = device

    @model_validator(mode="before")
    def default_logical_connectivity(cls, data):
        if not data.get("logical_connectivity", None):
            logical_connectivity = deepcopy(data["physical_connectivity"])
            data["logical_connectivity"] = logical_connectivity

        if not data.get("logical_connectivity_quality", None):
            logical_connectivity_quality = {}

            for q1_index, connected_qubits in data["logical_connectivity"].items():
                for q2_index in connected_qubits:
                    logical_connectivity_quality[(q1_index, q2_index)] = 1.0

            data["logical_connectivity_quality"] = logical_connectivity_quality

        return data

    @field_validator("physical_connectivity")
    def validate_physical_connectivity_symmetry(cls, physical_connectivity):
        for q, connected_qs in physical_connectivity.items():
            for connected_q in connected_qs:
                if (
                    physical_connectivity.get(connected_q, None)
                    and q not in physical_connectivity[connected_q]
                ):
                    raise ValueError(
                        f"The topology is not symmetric, qubit {q} not present in connected qubits of {connected_q}."
                    )
        return physical_connectivity

    @model_validator(mode="after")
    def validate_connectivity(self):
        # Check if all qubits exist in physical connectivity.
        if len(self.qubits) > 1:  # 1Q-systems do not have any connectivity.
            assert self.qubits.keys() == self.physical_connectivity.keys(), (
                f"Inconsistent qubit ids for {self.qubits} and {self.physical_connectivity}."
            )

        # Check if logical connectivity is subset of physical connectivity.
        for qubit_index in self.logical_connectivity:
            if (
                self.logical_connectivity[qubit_index]
                and not self.logical_connectivity[qubit_index]
                <= self.physical_connectivity[qubit_index]
            ):
                raise ValueError(
                    "Logical connectivity must be a subgraph of the physical connectivity."
                )

        # Check if qubit cross resonance (cancellation) pulse channels agree with the physical connectivity.
        physical_connectivities = {
            (src, tgt) for (src, tgts) in self.physical_connectivity.items() for tgt in tgts
        }
        cross_resonance_edges = {
            (src, chan.auxiliary_qubit)
            for src, qubit in self.qubits.items()
            for chan in qubit.cross_resonance_pulse_channels.values()
        }
        cross_cancellation_edges = {
            (src, chan.auxiliary_qubit)
            for src, qubit in self.qubits.items()
            for chan in qubit.cross_resonance_cancellation_pulse_channels.values()
        }

        assert cross_resonance_edges == cross_cancellation_edges, (
            "Cross resonance channels mismatch cross resonance cancellation channels."
        )
        assert physical_connectivities == cross_resonance_edges, (
            "Cross resonance channels mismatch physical connectivity."
        )
        assert physical_connectivities == cross_cancellation_edges, (
            "Cross resonance cancellation channels mismatch physical connectivity."
        )

        return self

    @model_validator(mode="after")
    def validate_connectivity_quality(self):
        for q1_index, q2_index in self.logical_connectivity_quality:
            if not self.logical_connectivity.get(q1_index, None):
                raise IndexError(
                    f"The coupling ({q1_index}, {q2_index}) is not present in `logical_connectivity`."
                )

        for q1_index, connected_qubits in self.logical_connectivity.items():
            for q2_index in connected_qubits:
                if (q1_index, q2_index) not in self.logical_connectivity_quality:
                    raise IndexError(
                        f"The coupling ({q1_index}, {q2_index}) is not present in `logical_connectivity_quality`."
                    )

        return self

    @model_validator(mode="after")
    def validate_error_mitigation(self):
        if self.error_mitigation.is_enabled and self.error_mitigation.qubits != list(
            self.qubits.keys()
        ):
            raise ValueError(
                "Please provide an error mitigation strategy for all qubits in the hardware model."
            )

        return self

    def __eq__(self, other: PhysicalHardwareModel) -> bool:
        base_eq = super().__eq__(other)

        s_qubits = list(self.qubits.values())
        o_qubits = list(other.qubits.values())
        if len(s_qubits) != len(o_qubits):
            return False

        for s, o in zip(s_qubits, o_qubits):
            if s != o:
                return False

        return base_eq

    @property
    def is_calibrated(self) -> bool:
        for qubit in self.qubits.values():
            if not qubit.is_calibrated:
                return False
        return True

    @property
    def number_of_qubits(self) -> int:
        return len(self.qubits)

    def qubit_with_index(self, index: int | QubitId) -> Qubit:
        return self.qubits[index]

    def index_of_qubit(self, qubit: Qubit) -> QubitId:
        """Returns the index of the given qubit in the hardware model."""
        return self._qubits_to_qubit_ids.get(qubit, None)

    def pulse_channel_with_id(self, id_: str):
        return self._ids_to_pulse_channels.get(id_, None)

    def physical_channel_with_id(self, id_: str):
        return self._ids_to_physical_channels.get(id_, None)

    def physical_channel_for_pulse_channel_id(self, id_: str):
        return self._pulse_channel_ids_to_physical_channel.get(id_, None)

    def device_for_pulse_channel_id(self, id_: str):
        return self._pulse_channel_ids_to_device.get(id_, None)

    def qubit_for_resonator(self, resonator: Resonator) -> Qubit:
        """Returns the qubit associated with the given resonator."""
        return self._resonators_to_qubits.get(resonator, None)

    @property
    def quantum_devices(self) -> list[Qubit, Resonator]:
        """
        Returns all quantum (an)harmonic oscillator devices
        in this hardware model as a list.
        """
        qubits = list(self.qubits.values())
        resonators = [qubit.resonator for qubit in qubits]
        return qubits + resonators


PydLogicalHardwareModel = LogicalHardwareModel
PydPhysicalHardwareModel = PhysicalHardwareModel



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.frontend.parsers.qasm.qasm2 import CloudQasmParser as CloudQasmParser
from qat.frontend.parsers.qasm.qasm2 import Qasm2Parser as Qasm2Parser
from qat.frontend.parsers.qasm.qasm2 import RestrictedQasm2Parser as RestrictedQasm2Parser
from qat.frontend.parsers.qasm.qasm3 import Qasm3Parser as Qasm3Parser
from qat.frontend.parsers.qir import QIRParser as QIRParser



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import abc

from qat.model.loaders.base import BaseModelLoader
from qat.purr.compiler.hardware_models import HardwareModel as LegacyHardwareModel


class BaseLegacyModelLoader(BaseModelLoader):
    @abc.abstractmethod
    def load(self) -> LegacyHardwareModel:
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from copy import deepcopy

from qat.model.device import (
    AcquirePulseChannel,
    CalibratableAcquire,
    CalibratablePulse,
    CrossResonanceCancellationPulseChannel,
    CrossResonancePulseChannel,
    DrivePulseChannel,
    FreqShiftPulseChannel,
    MeasurePulseChannel,
    PhysicalBaseband,
    Qubit,
    QubitPhysicalChannel,
    QubitPulseChannels,
    Resonator,
    ResonatorPhysicalChannel,
    ResonatorPulseChannels,
    SecondStatePulseChannel,
)
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.compiler.devices import ChannelType, PulseChannelView
from qat.purr.compiler.instructions import CustomPulse
from qat.utils.pydantic import FrozenDict


def convert_purr_echo_hw_to_pydantic(legacy_hw):
    new_qubits = {}
    logical_connectivity = defaultdict(set)

    for qubit in legacy_hw.qubits:
        # Add topology of qubit. Since the topology is not always stored both in
        # the `qubit_direction_couplings`, look into `coupled_qubits` as well.
        if couplings := getattr(legacy_hw, "qubit_direction_couplings", None):
            coupled_q_indices = [
                c.direction[1] for c in couplings if c.direction[0] == qubit.index
            ]
        else:
            coupled_q_indices = [q.index for q in qubit.coupled_qubits]
        logical_connectivity[qubit.index].update(coupled_q_indices)

        # Physical baseband
        phys_bb_q = qubit.physical_channel.baseband
        new_phys_bb_q = PhysicalBaseband(
            frequency=phys_bb_q.frequency, if_frequency=phys_bb_q.if_frequency
        )

        phys_bb_r = qubit.measure_device.physical_channel.baseband
        new_phys_bb_r = PhysicalBaseband(
            frequency=phys_bb_r.frequency, if_frequency=phys_bb_r.if_frequency
        )

        # Physical channel
        phys_channel_q = qubit.physical_channel
        new_phys_ch_q = QubitPhysicalChannel(
            baseband=new_phys_bb_q,
            sample_time=phys_channel_q.sample_time,
            block_size=phys_channel_q.block_size,
        )

        phys_channel_r = qubit.measure_device.physical_channel
        new_phys_ch_r = ResonatorPhysicalChannel(
            baseband=new_phys_bb_r,
            sample_time=phys_channel_r.sample_time,
            block_size=phys_channel_r.block_size,
            swap_readout_iq=getattr(phys_channel_r, "swap_readout_IQ", False),
        )

        # Resonator
        measure_pulse_channel = qubit.measure_device.get_pulse_channel(ChannelType.measure)
        pulse_measure = deepcopy(qubit.pulse_measure)
        pulse_measure.pop("shape")
        new_measure_pulse_channel = MeasurePulseChannel(
            frequency=measure_pulse_channel.frequency,
            imbalance=measure_pulse_channel.imbalance,
            scale=measure_pulse_channel.scale,
            fixed_if=measure_pulse_channel.fixed_if,
            phase_iq_offset=measure_pulse_channel.phase_offset,
            pulse=CalibratablePulse(**pulse_measure),
        )

        meas_acq = deepcopy(qubit.measure_acquire)
        if meas_acq.get("weights") is None:
            meas_acq["weights"] = []
        elif isinstance(meas_acq.get("weights"), CustomPulse):
            meas_acq["weights"] = meas_acq["weights"].samples

        acquire_pulse_channel = qubit.measure_device.get_pulse_channel(ChannelType.acquire)
        new_acquire_pulse_channel = AcquirePulseChannel(
            frequency=acquire_pulse_channel.frequency,
            imbalance=acquire_pulse_channel.imbalance,
            scale=acquire_pulse_channel.scale,
            fixed_if=acquire_pulse_channel.fixed_if,
            phase_iq_offset=acquire_pulse_channel.phase_offset,
            acquire=CalibratableAcquire(**meas_acq),
        )

        new_res_pulse_channels = ResonatorPulseChannels(
            measure=new_measure_pulse_channel, acquire=new_acquire_pulse_channel
        )

        resonator = Resonator(
            physical_channel=new_phys_ch_r,
            pulse_channels=new_res_pulse_channels,
        )

        # Qubit pulse channels
        drive_pulse_channel = qubit.get_pulse_channel(ChannelType.drive)
        pulse_hw_x_pi_2 = deepcopy(qubit.pulse_hw_x_pi_2)
        pulse_hw_x_pi_2.pop("shape")
        new_drive_pulse_channel = DrivePulseChannel(
            frequency=drive_pulse_channel.frequency,
            imbalance=drive_pulse_channel.imbalance,
            scale=drive_pulse_channel.scale,
            fixed_if=drive_pulse_channel.fixed_if,
            phase_iq_offset=drive_pulse_channel.phase_offset,
            pulse=CalibratablePulse(**pulse_hw_x_pi_2),
        )

        try:
            freqshift_pulse_channel = qubit.get_pulse_channel(ChannelType.freq_shift)
            new_freqshift_pulse_channel = FreqShiftPulseChannel(
                frequency=freqshift_pulse_channel.frequency,
                imbalance=freqshift_pulse_channel.imbalance,
                scale=freqshift_pulse_channel.scale,
                fixed_if=freqshift_pulse_channel.fixed_if,
                phase_iq_offset=freqshift_pulse_channel.phase_offset,
                active=freqshift_pulse_channel.active,
                amp=freqshift_pulse_channel.amp,
                phase=getattr(freqshift_pulse_channel.pulse_channel, "phase", 0.0),
            )
        except KeyError:
            new_freqshift_pulse_channel = FreqShiftPulseChannel()

        try:
            secondstate_pulse_channel = qubit.get_pulse_channel(ChannelType.second_state)
            new_secondstate_pulse_channel = SecondStatePulseChannel(
                frequency=secondstate_pulse_channel.frequency,
                imbalance=secondstate_pulse_channel.imbalance,
                scale=secondstate_pulse_channel.scale,
                fixed_if=secondstate_pulse_channel.fixed_if,
                phase_iq_offset=secondstate_pulse_channel.phase_offset,
            )
        except KeyError:
            new_secondstate_pulse_channel = SecondStatePulseChannel()

        new_cross_resonance_pulse_channels = {}
        new_cross_resonance_cancellation_pulse_channels = {}

        for pulse_channel in qubit.pulse_channels.values():
            if (
                isinstance(pulse_channel, PulseChannelView)
                and pulse_channel.auxiliary_devices
            ):
                aux_qubit = pulse_channel.auxiliary_devices[0].index

                if pulse_channel.channel_type == ChannelType.cross_resonance:
                    new_cr_pulse_channel = CrossResonancePulseChannel(
                        auxiliary_qubit=aux_qubit,
                        frequency=pulse_channel.frequency,
                        imbalance=pulse_channel.imbalance,
                        scale=pulse_channel.scale,
                        fixed_if=pulse_channel.fixed_if,
                        phase_iq_offset=pulse_channel.phase_offset,
                    )
                    new_cross_resonance_pulse_channels[aux_qubit] = new_cr_pulse_channel

                elif pulse_channel.channel_type == ChannelType.cross_resonance_cancellation:
                    new_crc_pulse_channel = CrossResonanceCancellationPulseChannel(
                        auxiliary_qubit=aux_qubit,
                        frequency=pulse_channel.frequency,
                        imbalance=pulse_channel.imbalance,
                        scale=pulse_channel.scale,
                        fixed_if=pulse_channel.fixed_if,
                        phase_iq_offset=pulse_channel.phase_offset,
                    )
                    new_cross_resonance_cancellation_pulse_channels[aux_qubit] = (
                        new_crc_pulse_channel
                    )

        new_cross_resonance_pulse_channels = FrozenDict(new_cross_resonance_pulse_channels)

        new_cross_resonance_cancellation_pulse_channels = FrozenDict(
            new_cross_resonance_cancellation_pulse_channels
        )

        new_qubit_pulse_channels = QubitPulseChannels(
            drive=new_drive_pulse_channel,
            freq_shift=new_freqshift_pulse_channel,
            second_state=new_secondstate_pulse_channel,
            cross_resonance_channels=new_cross_resonance_pulse_channels,
            cross_resonance_cancellation_channels=new_cross_resonance_cancellation_pulse_channels,
        )

        new_qubit = Qubit(
            physical_channel=new_phys_ch_q,
            pulse_channels=new_qubit_pulse_channels,
            resonator=resonator,
        )
        new_qubits[qubit.index] = new_qubit

    new_qubits = FrozenDict(new_qubits)
    physical_connectivity = defaultdict(set)
    for q1_index in logical_connectivity:
        for q2_index in logical_connectivity[q1_index]:
            physical_connectivity[q1_index].add(q2_index)
            physical_connectivity[q2_index].add(q1_index)

    if legacy_hw.qubit_direction_couplings:
        logical_connectivity_quality = {}
        for coupling in legacy_hw.qubit_direction_couplings:
            logical_connectivity_quality[coupling.direction] = (
                coupling.quality / 100.0
            )  # We assume that the quality is in [0, 100] in the legacy hw model.
    else:
        logical_connectivity_quality = None

    new_hw = PhysicalHardwareModel(
        logical_connectivity=logical_connectivity,
        logical_connectivity_quality=logical_connectivity_quality,
        physical_connectivity=physical_connectivity,
        qubits=new_qubits,
    )

    return new_hw
